/*  ISAM database library, by Ron Ruffin, copyright 1989 Excelsior Systems,  */
/*  All rights reserved.  This library is intended to serve as a programmer  */
/*  interface to the ISAM function library from Mix Software.  The latter    */
/*  library requires that a great deal of attention be focused on pointers   */
/*  and variables which are not of immediate concern to the programmer.  As  */
/*  much as was possible, we have confined such references to this library,  */
/*  to enable the programmer the freedom to focus on the needs of the        */
/*  application.  It should be noted that blank_line is borrowed from its    */
/*  rightful place in a library of standard RAR functions, until such a      */
/*  library can be created.  Also, this library does not address the need    */
/*  for field array manipulation.  Array pointers will need to be set before */
/*  writes, and after reads.                                                 */
/*$no trace*/

#include <stdio.h>
/*                             "standard  I/O library"                       */

#include <graphics.h>
#include <conio.h>
/*                             "console   I/O library"                       */

#include <string.h>
/*                             "string function library"                     */

#include "isam.h"
#include "rarkeys.c"

#define KEY_ONLY 2
/*    I edit_array field_type, this is the value which tells the next-prev
      to stop only on key fields.     */

#define KEY_YES 1
/*    In edit_array field_type, this is the value which tells the next-prev
      that the current screen coordinates represent a key field, and to stop
      on whichever field they land on next, without regard to whether or not
      it is a key field.    */

#define NON_KEY 0
/*    In edit_array field_type, this is the value which tells the next-prev
      that the current screen coordinates DO NOT represent a key field, and
      to only stop on non-key fields.   */

#define NUMERIC 2
/*    In edit_array field_type, this is the value which tells the edit_field
      function that the current screen coordinates represent a numeric field.
      The next_prev functions will also consider this to be a non-key
      field.  */
#pragma library "isam.lib"
#pragma library "cbt.lib"


int isam_status;
/*    Status variable, for use by calling program.  This variable will show
      whether the function called was successful.  The values are taken from
      the Mix Software return values.  OK means no problem occurred.
      NOTFOUND means a search was unsuccessful.  ERROR means some error
      occurred within the function.  */

char *fields[20];
/*           Array of pointers, to ISAM data fields on disk.                 */

char *key_fields[20];
/*       Array of pointers, to ISAM key data fields on disk.                 */

char buffer[860];
/*           Input buffer.                                                   */

int edit_array [20] [7];
/*   Array containing information about each field for editing.  The first
     dimension indicates the field number in the order the fields were
     displayed.  The second dimension contains information about each field.
     Element [0] = field start row.  Element [1] = field start column.
     Element [2] = field length.  Element [3] = field address.
     Element [4] = field type.  This information can be recalled for each
     field, using the field number.  Element [5] = number of digits before
     decimal (if field type = NUMERIC)  Element [6] = number of digits after
     decimal (if field type = NUMERIC).  */

int current_edit_field = 0;
/*  Current subscript pointer within edit_array.  Used by display_field,
    select_field, select_nk_field and clear_screen.  Allows calls to
    edit_field to access appropriate fields by field number in the order
    they were displayed, or to access the most recent field displayed. */

int last_field = 0;
/*  Holds the number of the most recent field displayed.  current_edit_field
    will be set from this variable, whenever it needs the most recent field
    number.  Whenever a new field is displayed, this variable is incremented.
    Whenever the screen is cleared, this field is set to 0.    */

char sf_action;
/*  Switch to determine whether the select-edit loop should edit, move back or
    forward one field, or exit.  'e' = edit, 'n' = next, 'p' = previous,
    'x' = exit.      */



isam_create(Db_Obj **database_handle, char *database_name,
    char *record_format[], Index_Obj **index_handle,
        char *index_name, char *index_format[])

/*  This function creates a database with filenames database_name.DB and     */
/*  database_name.IDX.  It gives these files the database handle             */
/*  database_handle, and structures the database to record_format, which     */
/*  must hold the names of the fields before this function is called.        */
/*  Note:  the database_handle and index_handle fields should be referenced  */
/*  by the call to this function, as addresses.                              */
{
    char key;

/*  Check to see if database exists already.  If it does, give the user the
    option of overwriting the existing database.                            */

    isam_silent_open(database_handle, database_name);
    if (isam_status != ERROR) {
        isam_close(*database_handle);
OVERWRITE_CHK:
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        printf("  Database already exists.  Overwrite? (Y/N)");
        key = getch();
        if (((key == 'N') || (key == 'n')) || (key == 13)) {
            printf ("Aborting file creation.");
            isam_status = ERROR;
            return;
        };
        if ((key != 'Y') && (key != 'y')) goto OVERWRITE_CHK;
    };

    isam_status = OK;
    *database_handle = icreate_db(database_name, 0, record_format);
    if (database_handle == NULL) {
        isam_status = ERROR;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        iprterr();
        printf("  Error creating database.  Press any key to continue.");
        getch();
        return;
    };
    isam_index(*database_handle, index_name, index_format);
    isam_index_handle(*database_handle, index_handle, index_name);
}








isam_index(Db_Obj *db_handle, char *idx_name, char *idx_format[])
/*  This function creates an index for the database represented by           */
/*  db_handle.  The index will be the field indicated by idx_name  The       */
/*  physical data field or fields which this index will represent, are shown */
/*  in idx_format.                                                           */
{
    int index_status;
    isam_status = OK;
    index_status = imkindex(db_handle, idx_name, idx_format);
    if (index_status == ERROR) {
        isam_status = ERROR;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        iprterr();
        printf(" Error creating index.  Press any key to continue.");
        getch();
    };
}


isam_index_handle(Db_Obj *db_handle, Index_Obj **idx_handle,
    char *idx_name)
/*  This function creates an index handle for the database represented by    */
/*  db_name, and the index represented by idx_name (which must be identical  */
/*  to the index name used in calling the isam_index function which created  */
/*  the index.  The index handle value is then returned to the calling       */
/*  function.                                                                */
{
    isam_status = OK;
    *idx_handle = ihandle(db_handle, idx_name);
    if (*idx_handle == NULL) {
        isam_status = ERROR;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        iprterr();
        printf(" Error creating index handle.  Press any key to continue.");
        getch();
    };
}




isam_save(Db_Obj *database_handle, Index_Obj *index_handle)

/*  This function saves a record to the database indicated by                */
/*  database_handle.  This function presumes that the field pointer array    */
/*  fields[] has been initialized to hold the addresses of the data items    */
/*  to be saved.                                                             */

{

    int record_status;
    isam_status = OK;
    record_status = ifindrec(database_handle, index_handle, fields);
    if (record_status == NOTFOUND) {
        record_status = iaddrec(database_handle, index_handle, fields);
        return;
    };
    if (record_status == FOUND) {
        record_status = imodrec(database_handle, index_handle, fields);
        return;
    };
    if (record_status == ERROR) {
        isam_status = ERROR;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        iprterr();
        printf("  Error saving current record.  Press any key to continue.");
        getch();
        return;
    };

}

isam_delete(Db_Obj *database_handle, Index_Obj *index_handle)

/*  This function deletes a record from the database indicated by            */
/*  database_handle.  This function presumes that the field pointer array    */
/*  fields[] has been initialized to hold the addresses of the data items    */
/*  to be deleted.                                                           */

{

    int record_status;
    isam_status = OK;
    record_status = ifindrec(database_handle, index_handle, fields);
    if (record_status == FOUND) {
        record_status = idelrec(database_handle, index_handle);
        return;
    };
    if (record_status == ERROR) {
        isam_status = ERROR;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        iprterr();
        printf("  Error deleting current record.  Press any key to continue.");
        getch();
        return;
    };

}


isam_first(Db_Obj *database_handle, Index_Obj *index_handle)
/*  This function locates the first record in the database represented by    */
/*  database_handle, using the index represented by index_handle.  Before    */
/*  this function is called, the field pointers should be set to point to    */
/*  their respective data fields.  After this function is completed, the     */
/*  data should be copied from the buffer (using the field pointers) to the  */
/*  intended data fields.                                                    */
{
    Rec_Len record_length;
    int record_status;
    isam_status = OK;
    record_status = ifindhead(database_handle, index_handle);
    if (record_status == EOI) {
        isam_status = EOI;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        printf("     Database is empty.  Press any key to continue.");
        getch();
        return;
    };
    if (record_status == ERROR) {
        isam_status = ERROR;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        iprterr();
        printf("     Error getting first record.  Press any key to continue.");
        getch();
        return;
    };
    record_status =
        igetreclen(database_handle, index_handle, &record_length);
    record_status =
        igetrec(database_handle, index_handle,
            fields, buffer, record_length);
}

isam_last(Db_Obj *database_handle, Index_Obj *index_handle)
/*  This function locates the last record in the database represented by     */
/*  database_handle, using the index represented by index_handle.  Before    */
/*  this function is called, the field pointers should be set to point to    */
/*  their respective data fields.  After this function is completed, the     */
/*  data should be copied from the buffer (using the field pointers) to the  */
/*  intended data fields.                                                    */
{
    Rec_Len record_length;
    int record_status;
    isam_status = OK;
    record_status = ifindtail(database_handle, index_handle);
    if (record_status == BOI) {
        isam_status = BOI;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        printf("     Database is empty.  Press any key to continue.");
        getch();
        return;
    };
    if (record_status == ERROR) {
        isam_status = ERROR;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        iprterr();
        printf("     Error getting last record.  Press any key to continue.");
        getch();
        return;
    };
    record_status =
        igetreclen(database_handle, index_handle, &record_length);
    record_status =
        igetrec(database_handle, index_handle,
            fields, buffer, record_length);
}

isam_next(Db_Obj *database_handle, Index_Obj *index_handle)
/*  This function locates the next record in the database represented by     */
/*  database_handle, using the index represented by index_handle.  Before    */
/*  this function is called, the field pointers should be set to point to    */
/*  their respective data fields.  After this function is completed, the     */
/*  data should be copied from the buffer (using the field pointers) to the  */
/*  intended data fields.                                                    */
{
    Rec_Len record_length;
    int record_status;
    isam_status = OK;
    record_status = ifindnext(database_handle, index_handle);
    if (record_status == EOI) {
        isam_status = EOI;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        printf("     Last record already shown.  Press any key to continue.");
        getch();
        return;
    };
    if (record_status == ERROR) {
        isam_status = ERROR;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        iprterr();
        printf("     Error finding next record.  Press any key to continue.");
        getch();
        return;
    };
    record_status =
        igetreclen(database_handle, index_handle, &record_length);
    if (record_status == ERROR) {
        isam_status = ERROR;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        iprterr();
        printf("     Error checking next record.  Press any key to continue.");
        getch();
        return;
    };
    record_status =
        igetrec(database_handle, index_handle,
            fields, buffer, record_length);
    if (record_status == ERROR) {
        isam_status = ERROR;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        iprterr();
        printf("     Error getting next record.  Press any key to continue.");
        getch();
        return;
    };
}




isam_silent_next(Db_Obj *database_handle, Index_Obj *index_handle)
/*  This function locates the next record in the database represented by     */
/*  database_handle, using the index represented by index_handle.  Before    */
/*  this function is called, the field pointers should be set to point to    */
/*  their respective data fields.  After this function is completed, the     */
/*  data should be copied from the buffer (using the field pointers) to the  */
/*  intended data fields.  This function does not display a message to user. */
{
    Rec_Len record_length;
    int record_status;
    isam_status = OK;
    record_status = ifindnext(database_handle, index_handle);
    if (record_status == EOI) {
        isam_status = EOI;
        return;
    };
    if (record_status == ERROR) {
        isam_status = ERROR;
        return;
    };
    record_status =
        igetreclen(database_handle, index_handle, &record_length);
    if (record_status == ERROR) {
        isam_status = ERROR;
        return;
    };
    record_status =
        igetrec(database_handle, index_handle,
            fields, buffer, record_length);
    if (record_status == ERROR) {
        isam_status = ERROR;
        return;
    };
}



isam_prev(Db_Obj *database_handle, Index_Obj *index_handle)
/*  This function locates the previous record in the database represented by */
/*  database_handle, using the index represented by index_handle.  Before    */
/*  this function is called, the field pointers should be set to point to    */
/*  their respective data fields.  After this function is completed, the     */
/*  data should be copied from the buffer (using the field pointers) to the  */
/*  intended data fields.                                                    */
{
    Rec_Len record_length;
    int record_status;
    isam_status = OK;
    record_status = ifindprev(database_handle, index_handle);
    if (record_status == BOI) {
        isam_status = BOI;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        printf("     First record already shown.  Press any key to continue.");
        getch();
        return;
    };
    if (record_status == ERROR) {
        isam_status = ERROR;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        iprterr();
        printf("     Error getting previous record.  Press any key to \
continue.");
        getch();
        return;
    };
    record_status =
        igetreclen(database_handle, index_handle, &record_length);
    record_status =
        igetrec(database_handle, index_handle,
            fields, buffer, record_length);
}


isam_open(Db_Obj **database_handle, char *database_name,
    Index_Obj **index_handle, char *index_name)
/*  This function opens the existing database with filename database_name,   */
/*  and assigns it the handle represented by database_handle.  It then       */
/*  finds the index index_name, and assigns it the handle index_handle.      */
/*  Note:  the database_handle and index_handle fields should be referenced  */
/*  by the call to this function, as addresses.                              */
{
    isam_status = OK;
    *database_handle = iopen_db(database_name);
    if (*database_handle == NULL) {
        isam_status = ERROR;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        iprterr();
        printf("  Error opening database.  Press any key to continue.");
        getch();
        return;
    };
    isam_index_handle(*database_handle, index_handle, index_name);
}



isam_silent_open(Db_Obj **database_handle, char *database_name)
/*  This function opens the existing database with filename database_name,   */
/*  and assigns it the handle represented by database_handle.  Unlike        */
/*  isam_open, it does not look for an index or assign an index handle.      */
/*  This function is only to be used to determine whether a database file    */
/*  already exists.  The database handle should be passed as an address.     */
{
    isam_status = OK;
    *database_handle = iopen_db(database_name);
    if (*database_handle == NULL) {
        isam_status = ERROR;
        return;
    };
}


isam_search(Db_Obj *database_handle, Index_Obj *index_handle,
    char *key_value)
/*  This function searches the database indicated by database_handle, using  */
/*  the index indicated by index_handle, for the key value specified in      */
/*  key_value.  It searches the indicated index of the database, for the     */
/*  specified key value, and retrieves any matching record.  After this      */
/*  function is complete, if the search was successful, the calling program  */
/*  must transfer the data from the fields[] array, to the data fields used  */
/*  by the program.                                                          */

{
    Rec_Len record_length;
    int record_status;
    isam_status = OK;
    empty_key_fields();
/*  blank_lines_after(24);
    poscurs(24, 0);
    printf("    Looking for search term \"%s\".", key_value);     */
    *(key_fields + 0) = key_value;
    *(key_fields + 1) = NULL;

    record_status =
        ifindkey(database_handle, index_handle, key_fields);
    if (record_status == NOTFOUND) {
        isam_status = NOTFOUND;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        printf("       Search term not found.  Press any key to continue");
        getch();
        return;
    };
    if (record_status == FOUND)  {
        record_status =
            igetreclen(database_handle, index_handle, &record_length);
        record_status =
            igetrec(database_handle, index_handle,
                fields, buffer, record_length);
    };
    if (record_status == ERROR) {
        isam_status = ERROR;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        iprterr();
        printf("  Error searching by key.  Press any key to continue.");
        getch();
    };

}

isam_search_5(Db_Obj *database_handle, Index_Obj *index_handle,
    char *key_value_1, char *key_value_2, char *key_value_3,
        char *key_value_4, char *key_value_5)
/*  This function searches the database indicated by database_handle, using  */
/*  the index indicated by index_handle, for the key value specified in      */
/*  key_value.  It searches the indicated index of the database, for the     */
/*  specified key value, and retrieves any matching record.  After this      */
/*  function is complete, if the search was successful, the calling program  */
/*  must transfer the data from the fields[] array, to the data fields used  */
/*  by the program.  Note:  This function is virtually identical to          */
/*  isam_search.  The only difference is the number of concurent key field   */
/*  value arguments.  Here the user can pass up to five key values.  Each    */
/*  key field represented here is less significant than the preceding field. */
/*  key_value_1 is more significant than key_value_2, and so on.             */

{
    Rec_Len record_length;
    int record_status;
    isam_status = OK;
    empty_key_fields();
/*  blank_lines_after(22);
    poscurs(22, 0);
    printf("    Looking for search terms.\n");
    printf("   %s - %s - %s - %s - %s", key_value_1, key_value_2, key_value_3,
        key_value_4, key_value_5);      */
    *(key_fields + 0) = key_value_1;
    *(key_fields + 1) = key_value_2;
    *(key_fields + 2) = key_value_3;
    *(key_fields + 3) = key_value_4;
    *(key_fields + 4) = key_value_5;
    *(key_fields + 5) = NULL;

    record_status =
        ifindkey(database_handle, index_handle, key_fields);
    if (record_status == NOTFOUND) {
        isam_status = NOTFOUND;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        printf("       Search terms not found.  Press any key to continue");
        getch();
        return;
    };
    if (record_status == FOUND)  {
        record_status =
            igetreclen(database_handle, index_handle, &record_length);
        record_status =
            igetrec(database_handle, index_handle,
                fields, buffer, record_length);
    };
    if (record_status == ERROR) {
        isam_status = ERROR;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        iprterr();
        printf("  Error searching by key.  Press any key to continue.");
        getch();
    };

}


isam_close(Db_Obj *db_handle)
{
    int file_status;
    isam_status = OK;
    file_status = iclose_db(db_handle);
    if (file_status == ERROR) {
        isam_status = ERROR;
        blank_line(FUNCTION_LINE);
        poscurs(FUNCTION_LINE, 0);
        iprterr();
        printf("  Error closing database.  Press any key to continue.");
        getch();
    };
}


blank_line(row_number)
/*       Move cursor to line # row_number, then print all blanks there.     */
{
    int column_number = 0;
    poscurs(row_number,0);
    while (column_number < 80) {
        printf(" ");
        ++column_number;
    };
}

blank_lines_after(row_number)
/*       Blank all lines from row_number to the bottom of the screen        */
{
    int current_row;
    for (current_row = row_number; current_row <= 25; current_row++) {
        blank_line(current_row);
    };
}


/*  Screen Edit Function Library, by Ron Ruffin.
    Copyright 1989 Excelsior Systems.  This library contains functions for
    full-screen editing.  With the aid of these functions, the programmer can
    set up screen displays, select data fields, and edit data fields directly
    on-screen, using the standard editing keys.                             */




edit_screen_data(field_type)
/*  Until the user is ready to exit, select fields to edit, and edit them.  */
{
    char quit_edit;

    sf_action = 'e';
    current_edit_field = last_field;
    move_to_next_field(field_type);
    curson();
 EDIT_TOP: quit_edit = select_field(field_type);
    if (quit_edit == 'Y') {
        current_edit_field = last_field;
        cursoff();
        return;
    };
    edit_field(current_edit_field);
    goto EDIT_TOP;
}



char select_field(field_type)
/*       Select field, then tell calling function whether or not to exit.   */
{
    int current_row;
    int current_column;


    if (last_field == 1000) return;
    current_row = edit_array [current_edit_field] [0];
    current_column = edit_array [current_edit_field] [1];
    poscurs(current_row, current_column);

    if (sf_action == 'e') return 'N';
    if (sf_action == 'x') return 'Y';
    if (sf_action == 'p') {
        move_back_one_field(field_type);
        sf_action = 'e';
        return 'N';
    };
    if (sf_action == 'n') {
        move_to_next_field(field_type);
        sf_action = 'e';
        return 'N';
    };

}



edit_field(field_number)
/*  This function performs a full screen edit on the data field indicated by
    field_number.  This number is applied to the first element of the
    edit_array, and the pertinent information about the data field is
    obtained from the second dimension of this array.  The user may use the
    editing keys to move the cursor about the field and to make change to it.
    When the ESC key is pressed, the data is stored in the field address. */
{
    char function1[81] = "ESCexit           Use editing keys.";
    int field_start_row;
    int field_start_col;
    int field_length;
    int field_address;
    int field_end_col;
    int field_end_row;
    field_start_row = edit_array [field_number] [0];
    field_start_col = edit_array [field_number] [1];
    field_length = edit_array [field_number] [2];
    field_address = edit_array [field_number] [3];
    determine_end_position(field_start_row, field_start_col, field_length,
        &field_end_row, &field_end_col);
    blank_line(FUNCTION_LINE);
    poscurs(FUNCTION_LINE,0);
    printf("%s",function1);
    poscurs(field_start_row, field_start_col);
EDIT_KEY_CHECK:  key = getch();
    if (key == ESC) {
        copy_edit(field_address, field_start_row, field_start_col,
            field_length);
        sf_action = 'x';
        return;
    };
    if ((key <=0x7E) && (key >= 0x20)) {
        if ((cursrow() == field_end_row) && (curscol() == field_end_col)) {
            printf("%c",key);
            left (field_start_row, field_start_col);
            goto EDIT_KEY_CHECK;
        };
        if ((cursrow() != field_end_row) || (curscol() != field_end_col)) {
            printf("%c",key);
            goto EDIT_KEY_CHECK;
        };
    };
    if (key == BACKSPACE) {
        backspace(field_start_row, field_start_col, field_length);
        goto EDIT_KEY_CHECK;
    };
    if ((key == TAB) || (key == ENTER)) {
        copy_edit(field_address, field_start_row, field_start_col,
            field_length);
        sf_action = 'n';
        return;
    };
    if (key == 0x00) {
        key = getch();
        if (key == INSERT) {
            insert(field_start_row, field_start_col, field_length);
            goto EDIT_KEY_CHECK;
        };
        if (key == DELETE) {
            delete(field_start_row, field_start_col, field_length);
            goto EDIT_KEY_CHECK;
        };
        if (key == UP) {
            up(field_start_row, field_start_col);
            goto EDIT_KEY_CHECK;
        };
        if (key == DOWN) {
            down(field_end_row, field_end_col);
            goto EDIT_KEY_CHECK;
        };
        if (key == RIGHT) {
            right(field_end_row, field_end_col);
            goto EDIT_KEY_CHECK;
        };
        if (key == LEFT) {
            left(field_start_row, field_start_col);
            goto EDIT_KEY_CHECK;
        };
        if (key == BACKTAB) {
            copy_edit(field_address, field_start_row, field_start_col,
                field_length);
            sf_action = 'p';
            return;
        };
    };
    goto EDIT_KEY_CHECK;
}









determine_end_position(start_row, start_column, length, end_row_address,
    end_column_address)
{
    int current_position;
    int *end_row;
    int *end_column;
    end_row = end_row_address;
    end_column = end_column_address;
    *end_row = start_row;
    *end_column = start_column;
    for (current_position = 1; current_position < length;
        current_position++) {
            ++*end_column;
            if (*end_column == 80) {
                *end_column = 0;
                ++*end_row;
            };
    };

}


copy_edit(field_address, start_row, start_col, field_length)
/*        Copy each screen editing field to its corresponding "scrn_" field.*/
{
    int end_row;
    int end_col;
    char *field_position;
    determine_end_position(start_row, start_col, field_length,
        &end_row, &end_col);
    field_position = field_address;
    poscurs(start_row, start_col);
    while ((cursrow() != end_row) || (curscol() != end_col)) {
        *field_position = readch();
        ++field_position;
        right(end_row, end_col);
    };
    ++field_position;
    *field_position = 0;
    strupr(field_address);
}



right(end_row, end_col)
/*         Move cursor one space to the right, or rollover to next line.    */
{
    int new_cursor_row;
    int new_cursor_col;
    if (curscol() < 79) {
        new_cursor_col = curscol() + 1;
        new_cursor_row = cursrow();
    };
    if (curscol() >= 79) {
        new_cursor_col = 0;
        new_cursor_row = cursrow() + 1;
    };
    if (new_cursor_row <= end_row) {
        if ((new_cursor_col <= end_col) || (new_cursor_row < end_row))
        poscurs(new_cursor_row, new_cursor_col);
    };

}



left(start_row, start_col)
/*   Move cursor one space to the left, or rollback to previous line.       */
{
    int new_cursor_row;
    int new_cursor_col;
    if (curscol() > 0) {
        new_cursor_col = curscol() - 1;
        new_cursor_row = cursrow();
    };
    if (curscol() <= 0) {
        new_cursor_col = 79;
        new_cursor_row = cursrow() - 1;
    };
    if (new_cursor_row >= start_row) {
        if ((new_cursor_col >= start_col) || (new_cursor_row > start_row))
        poscurs(new_cursor_row, new_cursor_col);
    };

}



down(end_row, end_col)
/*                Move cursor down one space within field.                  */
{
    int new_cursor_row;
    int new_cursor_col;
    new_cursor_col = curscol();
    new_cursor_row = cursrow() + 1;
    if ((new_cursor_row < end_row) || ((new_cursor_col <= end_col)
        && (new_cursor_row == end_row)))
        poscurs(new_cursor_row, new_cursor_col);
}



up(start_row, start_col)
/*                Move cursor up one space within field.                    */
{
    int new_cursor_row;
    int new_cursor_col;
    new_cursor_col = curscol();
    new_cursor_row = cursrow() - 1;
    if ((new_cursor_row > start_row) || ((new_cursor_col >= start_col)
        && (new_cursor_row == start_row)))
        poscurs(new_cursor_row, new_cursor_col);
}



insert(start_row, start_col, field_length)
{
    int mark_row;
    int mark_col;
    int end_row;
    int end_col;
    determine_end_position(start_row, start_col, field_length,
        &end_row, &end_col);
    mark_row = cursrow();
    mark_col = curscol();
    poscurs(end_row, end_col);
    while ((cursrow() != mark_row) || (curscol() != mark_col)) {
        left(start_row, start_col);
        copy_right(start_row, start_col, end_row, end_col);
    };
    putch(' ');
    left(start_row, start_col);

}


delete(start_row, start_col, field_length)
{
    int mark_row;
    int mark_col;
    int end_row;
    int end_col;
    determine_end_position(start_row, start_col, field_length,
        &end_row, &end_col);
    mark_row = cursrow();
    mark_col = curscol();
    while ((cursrow() != end_row) || (curscol() != end_col)) {
        right(end_row, end_col);
        copy_left(start_row, start_col, end_row, end_col);
    };
    putch(' ');
    poscurs(mark_row, mark_col);

}


copy_left(start_row, start_column, end_row, end_column)
/*  This function reads the character at the current cursor position, copies
    it one position to the left, then returns the cursor to its original
    position.       */
{
    char char_hold;
    char_hold = readch();
    left(start_row, start_column);
    putch(char_hold);
    left(start_row, start_column);
    right(end_row, end_column);
}


copy_right(start_row, start_column, end_row, end_column)
/*  This function reads the character at the current cursor position, copies
    it one position to the right, then returns the cursor to its original
    position.       */
{
    char char_hold;
    char_hold = readch();
    right(end_row, end_column);
    putch(char_hold);
    left(start_row, start_column);
    left(start_row, start_column);
}


backspace(start_row, start_col, field_length)
{
    if ((cursrow() == start_row) && (curscol() == start_col)) return;
    left(start_row, start_col);
    delete(start_row, start_col, field_length);

}



display_field(field_start_row, field_start_col, field_length,
    field_name_location, field_address, field_type)
/*  Display field_name at the screen position indicated by field_start_row
    and field_start_col.  Then add the cursor position and other field
    information to the edit_array, and toggle video to reverse.  Display
    the field contents at that location and toggle video back to normal.     */
{
    int char_pos;
    char *field_contents;
    char *field_name;
    field_contents = field_address;
    field_name = field_name_location;
    poscurs(field_start_row, field_start_col);
    printf("%s", field_name);
    add_field_entry(cursrow(), curscol(), field_length, field_address,
        field_type, 0, 0);
    toggle_video('r');
    for (char_pos = 0; char_pos < field_length; char_pos++) printf(" ");
    field_start_row = edit_array [last_field] [0];
    field_start_col = edit_array [last_field] [1];
    poscurs(field_start_row, field_start_col);
    printf("%s", field_contents);
    toggle_video('n');
}

toggle_video(char video_mode)
/*    This function sets the screen video to reverse if video_mode = 'r', or
      normal if video_mode = 'n'.  It should work regardless of whether the
      video board is a CGA or a Hercules.                                   */
{
/*  The following character strings, when printed, use ANSI to change the
    video attributes of all printed characters.                             */

    char *reverse_white = "\x1b[30;47m";
    char *reverse_green = "\x1b[30;42m";
    char *normal_white = "\x1b[0;37m";
    char *normal_green = "\x1b[0;32m";


    if (video_mode == 'r') {
        if ((getvmode() == 2) || (getvmode() == 7)) {
            printf("%s", reverse_white);
            return;
        };
        if (getvmode() == 3) {
            printf("%s", reverse_green);
            return;
        };
    };
    if (video_mode == 'n') {
        if ((getvmode() == 2) || (getvmode() == 7)) {
            printf("%s", normal_white);
            return;
        };
        if (getvmode() == 3) {
            printf("%s", normal_green);
            return;
        };
    };
}




display_and_edit(start_row, start_col, field_length, field_label,
    field_address, field_type, whole_digits, decimal_digits)
{
    display_numeric_field(start_row, start_col, field_length, field_label,
        field_address, field_type, whole_digits, decimal_digits);
    current_edit_field = last_field;
    edit_field(last_field);
}



add_field_entry(start_row, start_col, field_length, field_address,
    field_type, whole_digits, decimal_digits)
{
    if (last_field == 1000) last_field = 0;
    else ++last_field;
    edit_array [last_field] [0] = start_row;
    edit_array [last_field] [1] = start_col;
    edit_array [last_field] [2] = field_length;
    edit_array [last_field] [3] = field_address;
    edit_array [last_field] [4] = field_type;
    edit_array [last_field] [5] = whole_digits;
    edit_array [last_field] [6] = decimal_digits;

}


clear_edit_array()
{
    int x_element;
    for (x_element = 0; x_element < 20; x_element++) {
        edit_array [x_element] [0] = 500;
    };
}



empty_key_fields()
/*  This function empties the key_fields array, for use by the  key search  */
/*  functions.                                                              */
{
    int field_offset;
    for(field_offset = 0; field_offset < 20; field_offset++) {
        *(key_fields + field_offset) = NULL;
    };
}


move_to_next_field(field_type)
/*  This function moves the cursor to the beginning of the next field on the
    screen, using the edit_array values.  If the field_type not KEY_YES, any
    edit array 1st dimension entries containg a value of KEY_YES in the third
    column, will be skipped.     */
{

MOVE_NEXT_LOOP:
    if (current_edit_field == last_field) current_edit_field = 0;
    else ++current_edit_field;
    poscurs(edit_array [current_edit_field] [0],
        edit_array [current_edit_field] [1]);
    if (field_type == NON_KEY) {
        while (edit_array [current_edit_field] [4] == KEY_YES) {
            goto MOVE_NEXT_LOOP;
        };
    };
    if (field_type == KEY_ONLY) {
        while (edit_array [current_edit_field] [4] != KEY_YES) {
            goto MOVE_NEXT_LOOP;
        };
    };
}



move_back_one_field(field_type)
/*  This function moves the cursor to the beginning of the previous field on
    the screen, using the edit_array values.  If the field_type is NON-KEY, any
    edit array 1st dimension entries containg a value of KEY_YES in the third
    column, will be skipped.     */
{
MOVE_PREV_LOOP:
    if (current_edit_field == 0) current_edit_field = last_field;
    else --current_edit_field;
    poscurs(edit_array [current_edit_field] [0],
        edit_array [current_edit_field] [1]);
    if (field_type == NON_KEY) {
        while (edit_array [current_edit_field] [4] == KEY_YES) {
            goto MOVE_PREV_LOOP;
        };
    };
    if (field_type == KEY_ONLY) {
        while (edit_array [current_edit_field] [4] != KEY_YES) {
            goto MOVE_PREV_LOOP;
        };
    };
}



find_current_edit_field(start_row, start_col)
/*  This function determines the appropriate value of current_edit_field, given
    the start position of the field as coordinates.  */
{
    current_edit_field = 0;
    while (((edit_array [current_edit_field] [0] != start_row)
        || (edit_array [current_edit_field] [1] != start_col))
            && (current_edit_field < 20)) {
                ++current_edit_field;
    };
}


clear_screen()
{
    if ((getvmode() == 2) || (getvmode() == 7)) {
        clrscrn();
        return;
    };
    if (getvmode() == 3) clrscrn2(0x02);
    clear_edit_array();
    last_field = 1000;
}

display_numeric_field(field_start_row, field_start_col, field_length,
    field_name_location, field_address, field_type, whole_digits,
    decimal_digits)
/*  Display field_name at the screen position indicated by field_start_row
    and field_start_col.  Then add the cursor position and other field
    information to the edit_array, and toggle video to reverse.  Display
    the field contents, including decimal point if the field is numeric,
    then toggle video back to normal.  */
{
    int char_pos;
    char *field_contents;
    char *field_name;
    field_contents = field_address;
    field_name = field_name_location;
    poscurs(field_start_row, field_start_col);
    printf("%s", field_name);
    add_field_entry(cursrow(), curscol(), field_length, field_address,
        field_type, whole_digits, decimal_digits);
    toggle_video('r');
    for (char_pos = 0; char_pos < field_length; char_pos++) printf(" ");
    field_start_row = edit_array [last_field] [0];
    field_start_col = edit_array [last_field] [1];
    poscurs(field_start_row, field_start_col);
    printf("%s", field_contents);
    toggle_video('n');
}


function(char *function_key, char *function_label)
/*  This function composes a function key menu item, by displaying the
    function key in reverse video, followed by the key label in normal
    video.  This should make bottom-of-screen menus easier to code.     */
{
    toggle_video('r');
    printf(function_key);
    toggle_video('n');
    printf(function_label);

}

/*$trace*/
